/*
 * Copyright 1996 by Michel D. Grimminck
 *
 * All Rights Reserved
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * This file modified for Win32 and compressed memory-mapped files
 * by Harm Jetten, 1998.
 */

#ifndef _WIN32
#include <unistd.h>
#ifdef _POSIX_MAPPED_FILES
#include <sys/mman.h>
#endif
#else
#include <windows.h>
#include <direct.h>
#endif
#include <string.h>
#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <errno.h>
#include "var.h"
#include "functions.h"

#define DB_INDEX_FILE "endings"
#define DB_UNKNOWN 3
#define CAPTURE 100
#define QMAX 10000000
#define RAZ_PATH "databases/%s.raz"  /* mildly compressed "raw" files */
#define CPR_PATH "databases/%s.cpr"  /* better compressed excluding captures */
#define BLOCKSIZE 1024

char line[256];
int qsize[2],qmax[2];
char qname[2][20]={"/tmp/pos_queue0","/tmp/pos_queue1"};
FILE *qfile[2]={NULL,NULL};
unsigned char *database[4096];
int verified[4096];
int compressed[4096];
int loadeddb[4096];
int dbsize[4096];
#ifndef _WIN32
#ifdef _POSIX_MAPPED_FILES
int hfile[4096];
#endif
#else
HANDLE hfile[4096];
HANDLE hmap[4096];
#endif
unsigned int pow3[]={ 1,3,9,27,81 };
unsigned char *db_cptr;
int compr_bytex, compr_offset;
int curr_bx, curr_index, last_count;
unsigned char curr_byte, last_byte;
unsigned char *cpr_database[4096];
FILE *db_fp;
int save_blk[256];
int nr_pc;
int pc2code[8], code2pc[4];

/* 'database' points to all databases with a maximum of 7 pieces of each kind,
   or contains NULL if the database is not present.
   All databases are computed for 'white to move'. If black is to move
   the board is reversed.
*/

void log(char *str)
{
  int i;
  char tstr[80];
  FILE *logfile=fopen("log","a");
  if (logfile!=NULL) {
    time_t tm=time(NULL);
    strcpy(tstr,ctime(&tm));
    i=strlen(tstr);
    if (i>0&&tstr[i-1]=='\n')tstr[i-1]=0;
    if (i>1&&tstr[i-2]=='\r')tstr[i-2]=0;
    fprintf(logfile,"%s:\t%s",tstr,str);
    fclose(logfile);
  }
}

void init_queue_write(int q)
{
  pc2code[white|man]   = 0;
  pc2code[black|man]   = 1;
  pc2code[white|crown] = 2;
  pc2code[black|crown] = 3;
  if (q!=0 && q!=1) {printf("fatal: queue does not exist\n"); exit(1);}
  qsize[q]=0;
  qfile[q]=fopen(qname[q],"wb");
  if (qfile[q]==NULL) {printf("warning: can not open queue for writing\n"); return;}
}

void init_queue_read(int q)
{
  code2pc[0] = white|man;
  code2pc[1] = black|man;
  code2pc[2] = white|crown;
  code2pc[3] = black|crown;
  if (q!=0 && q!=1) {printf("fatal: queue does not exist\n"); exit(1);}
  if (qsize[q]<=0) {printf("fatal: attempting to open incomplete queue\n"); exit(1);}
  qfile[q]=fopen(qname[q],"rb");
  if (qfile[q]==NULL) {printf("fatal: can not open queue for reading\n"); exit(1);}
}

void add_to_queue(int q)
{
  int i;

  if (qfile[q]==NULL) return; /* queue not open (anymore) */
  if (qsize[q]>=qmax[q]) {
    printf("\nnote: maximum queue size reached: closing queue\n");
    fclose(qfile[q]);
    qfile[q]=NULL;
    qsize[q]=-1; /* queue incomplete: don't use it */
    return;
  }
  for(i=0;i<50;i++)
    if (board[map[i]]!=empty)
      fputc(i+(pc2code[board[map[i]]]<<6),qfile[q]);
  qsize[q]++;
}

static char* read_from_queue(int q)
{
  int i,p;
  static char local[93];

  if (qfile[q]==NULL) {printf("fatal: attempting to read from closed queue\n"); exit(1);}
  for(i=0;i<50;i++) local[map[i]]=empty;
  if (feof(qfile[q])) {
    fclose(qfile[q]);
    qfile[q]=NULL;
    return(local);
  }
  for(i=0;i<nr_pc;i++) {
    p=fgetc(qfile[q]);
    local[map[p&63]]=code2pc[p>>6];
  }
  return(local);
}

void close_queue(int q)
{
  if (qfile[q]!=NULL) fclose(qfile[q]);
  qfile[q]=NULL;
}

void swapqueues(int *q1,int *q2)
{
  /* if all went ok, q2 contains a valid queue: use this as input queue */
  *q1=1-*q1; *q2=1-*q2;
}

static char *database_name(int wman,int wcrown,int bman,int bcrown)
{
  static char name[40];
  int a=0,i;

  for(i=0;i<wcrown;i++) name[a++]='X';
  for(i=0;i<wman;i++) name[a++]='O';
  name[a++]='v';
  for(i=0;i<bcrown;i++) name[a++]='X';
  for(i=0;i<bman;i++) name[a++]='O';
  name[a]='\0';
  return(name);
}

int database_nr(int color,int wman,int wcrown,int bman,int bcrown)
/* returns the number of the requested database */
{
  int nr;

  if (color==white) nr=wman+8*wcrown+64*bman+512*bcrown;
  else nr=bman+8*bcrown+64*wman+512*wcrown;
  return(nr);
}

void database_pieces(int nr, int *wman, int *wcrown, int *bman, int *bcrown)
/* returns the number of pieces for database "nr", white to move */
{
  *wman=nr%8; *wcrown=(nr/8)%8; *bman=(nr/64)%8; *bcrown=(nr/512)%8;
}

void init_databases(void)
{
  int i;

  for(i=0;i<4096;i++) {
    database[i]=NULL;
#ifndef _WIN32
#ifdef _POSIX_MAPPED_FILES
    hfile[i]=-1;
#endif
#else
    hmap[i]=NULL;
    hfile[i]=INVALID_HANDLE_VALUE;
#endif
    verified[i]=false;
    compressed[i]=false;
  }
  init_index();
}

void close_databases(void)
/* close all dependent databases */
{
  int i;

  for(i=0;i<4096;i++) {
    if (compressed[i]==true) {
#ifndef _WIN32
#ifdef _POSIX_MAPPED_FILES
      munmap(database[i],dbsize[i]);
      close(hfile[i]);
      hfile[i]=-1;
#else
      free(database[i]);
#endif
#else
      UnmapViewOfFile(database[i]);
      CloseHandle(hmap[i]);
      CloseHandle(hfile[i]);
      hfile[i]=INVALID_HANDLE_VALUE;
      hmap[i]=NULL;
#endif
      database[i]=NULL;
      compressed[i]=false;
    }
  }
}

/* Map view of file to virtual memory */
#ifdef _WIN32
int load_database_memory_mapped(char *db_file, int nr)
{
  hfile[nr] = CreateFile(db_file, GENERIC_READ, FILE_SHARE_READ, NULL,
                         OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
  if (hfile[nr] == INVALID_HANDLE_VALUE)
    return false;
  hmap[nr] = CreateFileMapping(hfile[nr], NULL, PAGE_READONLY, 0, 0, db_file);
  if (hmap[nr] == NULL) {
    CloseHandle(hfile[nr]);
    hfile[nr] = INVALID_HANDLE_VALUE;
    return false;
  }
  database[nr] = MapViewOfFile(hmap[nr], FILE_MAP_READ, 0, 0, 0);
  if (database[nr] == NULL) {
    CloseHandle(hmap[nr]);
    hmap[nr] = NULL;
    CloseHandle(hfile[nr]);
    hfile[nr] = INVALID_HANDLE_VALUE;
    return false;
  }
  return true;
}
#else
#ifdef _POSIX_MAPPED_FILES
int load_database_memory_mapped(char *db_file, int nr)
{
  hfile[nr] = open(db_file,"rb");
  if (hfile[nr] == -1)
    return false;
  database[nr] = mmap(0, dbsize[nr], PROT_READ, MAP_SHARED, hfile[nr], 0);
  if (database[nr] == MAP_FAILED) {
    database[nr]=NULL;
    close(hfile[nr]);
    hfile[nr]=-1;
    return false;
  }
  return true;
}
#endif
#endif

/* retrieve database file status */
int my_stat(char *name,struct stat *stbuf)
{
  int result;

  result=stat(name,stbuf);
  if (result==-1) {
#ifndef _WIN32
    sprintf(line,"/usr/local/lib/dragon/%s",name);
    result=stat(line,stbuf);
#endif
#if 0                           /* for Win32 systems with small partitions: */
    sprintf(line,"F:/%s",name);
    strcpy(name,line);          /* "name" is in/out parameter */
    result=stat(name,stbuf);
    if (result==-1) {           /* try partition F:, then E: */
      name[0]='E';
      result=stat(name,stbuf);
    }
#endif
  }
  return result;
}

/* load the indicated database */
int load_database(int nr)
{
  char db_file[100],*id;
  FILE *db;
  int wman,wcrown,bman,bcrown;
  struct stat stbuf;
  int bytesize;

  database_pieces(nr,&wman,&wcrown,&bman,&bcrown);
  id=database_name(wman,wcrown,bman,bcrown);
  sprintf(db_file,RAZ_PATH,id);
  if (my_stat(db_file,&stbuf)==-1) {
    printf("error: database %s not found\n",db_file);
    return 0;
  }
  dbsize[nr]=bytesize=stbuf.st_size;
#if !defined(_WIN32) && !defined(_POSIX_MAPPED_FILES)
  database[nr]=(unsigned char *)malloc(sizeof(unsigned char)*bytesize);
  if (database[nr]==NULL) {
    /* swap space may be exhausted, close all and try again */
    printf("!");fflush(stdout);
    close_databases();
    database[nr]=(unsigned char *)malloc(sizeof(unsigned char)*bytesize);
    if (database[nr]==NULL) {
      printf("fatal: no memory for database\n");
      exit(1);
    }
  }
  db=my_fopen(db_file,"rb");
  if (db==NULL) {
    printf("error: database %s not found\n",db_file);
    free(database[nr]);
    database[nr]=NULL;
    return 0;
  }
  if (fread(database[nr],sizeof(unsigned char),bytesize,db)!=bytesize) {
    printf("error: database %s read failure\n",db_file);
    free(database[nr]);
    database[nr]=NULL;
    return 0;
  }
  fclose(db);
#else
  if (!load_database_memory_mapped(db_file, nr)) {
    /* virtual memory space may be exhausted, close all and try again */
    printf("!");fflush(stdout);
    close_databases();
    if (!load_database_memory_mapped(db_file, nr)) {
      printf("fatal: error opening database %s\n",db_file);
      exit(1);
    }
  }
#endif
  compressed[nr]=true;
  return bytesize;
}

int nextboard(int *list,int *type,int *constraint,int *min,int *max,int p)
{
  int try,stop=false;

  if (p<0) {printf("\n");return(true);}
  if (p==0) {printf(".");fflush(stdout);} /* progress marker */
  board[map[list[p]]]=empty;
  do {
    list[p]++;
  } while(board[map[list[p]]]!=empty && list[p]<max[p]);
  if (list[p]>=max[p]) {
  back:
    stop=nextboard(list,type,constraint,min,max,p-1);
    try=min[p];
    if (constraint[p]==1) try=list[p-1]+1;
    while(board[map[try]]!=empty && try<max[p]) try++;
    if (try>=max[p]) goto back;
    list[p]=try;
  }
  board[map[list[p]]]=type[p];
  return(stop);
}

int init_nextboard(int *list,int *type,int *constraint,int *min,int *max,int wman,int wcrown,int bman,int bcrown)
{
  int i,j,n;
  
  init_board();
  for(i=0;i<50;i++) board[map[i]]=empty;
  for(i=0;i<8;i++) list[i]=49;
  for(i=0;i<8;i++) constraint[i]=1;
  j=0;
  constraint[j]=0;
  for(i=0;i<bman;i++) {min[j]=0; max[j]=45; type[j++]=black|man;}
  constraint[j]=0;
  for(i=0;i<wman;i++) {min[j]=5; max[j]=50; type[j++]=white|man;}
  constraint[j]=0;
  for(i=0;i<bcrown;i++) {min[j]=0; max[j]=50; type[j++]=black|crown;}
  constraint[j]=0;
  for(i=0;i<wcrown;i++) {min[j]=0; max[j]=50; type[j++]=white|crown;}
  n=j;
  nextboard(list,type,constraint,min,max,n-1);
  return(n);
}

/* read from a compressed database */
int read_compr(int nr,INDEX index)
{
  unsigned char byte, *blkptr, *dbptr;
  int i,np,wman,wcrown,bman,bcrown;

  database_pieces(nr,&wman,&wcrown,&bman,&bcrown);
  np=wman+wcrown+bman+bcrown;
  dbptr = database[nr];
  blkptr=dbptr + (np>=6?4:3)*(index/BLOCKSIZE);
  dbptr += *blkptr++;           /* find start of segment */
  dbptr += *blkptr++ * 256;     /* note: is little-endian */
  dbptr += *blkptr++ * 65536;
  if (np>=6) dbptr += *blkptr * 16777216;
  i = index%BLOCKSIZE;
  do {
    byte = *dbptr++;
    if (byte <= 242) {
      i -= 5;
    } else if (byte <= 246) {
      if (byte == 246) i -= *dbptr++*5;
      else i -= (byte-241)*5;
      byte = 0;
    } else if (byte <= 250) {
      if (byte == 250) i -= *dbptr++*5;
      else i -= (byte-245)*5;
      byte = 121;
    } else if (byte <= 254) {
      if (byte == 254) i -= *dbptr++*5;
      else i -= (byte-249)*5;
      byte = 242;
    } else {
      i -= *dbptr++*5;
      byte = *dbptr++;
    }
  } while (i >= 0);
  return (byte/pow3[4+(i+1)%5])%3;
}

int read_value(int nr,INDEX index)
{
  if (compressed[nr]==true) return read_compr(nr,index);
  return (database[nr][index/4] >> (6-2*(index%4))) & 3;
}

int countvaluequick(int wman,int wcrown,int bman,int bcrown)
/* Prints result-counts of a database, and  */
/* returns the database's dominant result. */
/* NOTE: draw and unknown counts may be inaccurate */
/*       because of holes in the index function.   */
{
  INDEX index,pos_count,count[4]={0,0,0,0};
  int nr,result,mx,i;

  nr=database_nr(white,wman,wcrown,bman,bcrown);
  if (database[nr]==NULL)
    load_database(nr);    /* load on demand */
  if (database[nr]==NULL) {
    printf("error: database is not present\n");
    return 0;
  }
  pos_count=db_count(wman,wcrown,bman,bcrown);
  for(index=0;index<pos_count;index++)
    count[read_value(nr,index)]++;
  printf("approximate:  win %lu  draw %lu  loss %lu  unknown %lu\n",
         count[DB_WIN],count[DB_DRAW],count[DB_LOSE],count[DB_UNKNOWN]);
  count[DB_DRAW]+=count[DB_UNKNOWN];
  count[DB_UNKNOWN]=mx=result=0;
  for(i=0;i<4;i++)
    if (count[i]>mx) mx=count[i],result=i;
  return result;
}

/* write an index value to the database file */
void db_wr_index(int val)
{
  if (fwrite(&val,(nr_pc>=6?4:3),1,db_fp) != 1) { /* note: is little-endian */
    printf("fatal: compressed database index write error\n");
    exit(1);
  }
}

/* prepare writing a compressed database */
void db_start(char *db_file, int nr, INDEX pos_count)
{
  int i;

  db_fp=fopen(db_file,"wb");
  if (db_fp==NULL) {
    printf("fatal: can't open compressed database %s for writing\n",db_file);
    exit(1);
  }
  compr_bytex = curr_index = curr_byte = curr_bx = last_count = 0;
  compr_offset = ((pos_count+BLOCKSIZE-1)/BLOCKSIZE)*(nr_pc>=6?4:3);
  for (i=0; i<256; i++) save_blk[i]=0;
}

void db_wr_bytes(int count, int byte)
{
  if (count == 0) {
    return;
  } else if (count == 1) {
    db_cptr[compr_bytex++] = byte;
  } else if (byte == 0) {
    if      (count == 2) db_cptr[compr_bytex++] = 243;
    else if (count == 3) db_cptr[compr_bytex++] = 244;
    else if (count == 4) db_cptr[compr_bytex++] = 245;
    else db_cptr[compr_bytex++] = 246, db_cptr[compr_bytex++] = count;
  } else if (byte == 121) {
    if      (count == 2) db_cptr[compr_bytex++] = 247;
    else if (count == 3) db_cptr[compr_bytex++] = 248;
    else if (count == 4) db_cptr[compr_bytex++] = 249;
    else db_cptr[compr_bytex++] = 250, db_cptr[compr_bytex++] = count;
  } else if (byte == 242) {
    if      (count == 2) db_cptr[compr_bytex++] = 251;
    else if (count == 3) db_cptr[compr_bytex++] = 252;
    else if (count == 4) db_cptr[compr_bytex++] = 253;
    else db_cptr[compr_bytex++] = 254, db_cptr[compr_bytex++] = count;
  } else if (count <= 3) {
    while (count--) db_cptr[compr_bytex++] = byte;
  } else {
    db_cptr[compr_bytex++] = 255;     /* repeat marker */
    db_cptr[compr_bytex++] = count;   /* repeat count */
    db_cptr[compr_bytex++] = byte;
  }
}

/* wrap up current block */
void db_wrapupblock(void)
{
  if (curr_bx != 0) {   /* fill up last byte */
    while (curr_bx < 5)
      curr_byte += pow3[curr_bx++]*(curr_byte%3);
    if (last_count == 0) {
      last_count = 1;
      last_byte = curr_byte;
    } else if (last_byte == curr_byte) {
      last_count++;
    } else {
      db_wr_bytes(last_count, last_byte);
      last_count = 1;
      last_byte = curr_byte;
    }
  }
  if (last_count == (BLOCKSIZE+4)/5) { /* segment with all identical values */
    if (save_blk[last_byte] == 0) {
      save_blk[last_byte] = compr_offset+curr_index;
      db_wr_bytes(last_count, last_byte);
    }
    db_wr_index(save_blk[last_byte]);
  } else {
    db_wr_bytes(last_count, last_byte);
    db_wr_index(compr_offset+curr_index);
  }
  last_count = curr_byte = curr_bx = 0;
}

/* add a position to a compressed database */
void db_addpos(INDEX index, int value, int exact)
{
  if (index != 0 && index % BLOCKSIZE == 0) {
    db_wrapupblock();
    curr_index = compr_bytex;
  }
  if (!exact) {  /* value is don't-care, pick best one for good compression */
    if (curr_bx != 0) value = curr_byte%3;
    else if (last_count != 0) value = last_byte%3;
  }
  curr_byte += pow3[curr_bx++]*value;    
  if (curr_bx == 5) {
    if (last_count == 0) {
      last_count = 1;
      last_byte = curr_byte;
    } else if (last_byte == curr_byte) {
      last_count++;
    } else {
      db_wr_bytes(last_count, last_byte);
      last_count = 1;
      last_byte = curr_byte;
    }
    curr_byte = curr_bx = 0;
  }
}

/* finish up writing a compressed database */
void db_finish(void)
{
  db_wrapupblock();
  if (fwrite(db_cptr,sizeof(unsigned char),compr_bytex,db_fp) != compr_bytex) {
    printf("fatal: compressed database write error\n");
    exit(1);
  }
  fclose(db_fp);
}

/* compress a database, with capture exclusion */
void db_compress(int wman,int wcrown,int bman,int bcrown)
{
  INDEX index,nindex,j,pos_count;
  int nr,n,score,dom;
  int list[8],type[8],constraint[8],min[8],max[8];
  char db_file[100],*id;

  id=database_name(wman,wcrown,bman,bcrown);
  printf("db_compress %s\n",id);
  nr=database_nr(white,wman,wcrown,bman,bcrown);
  if (database[nr]==NULL)
    load_database(nr);    /* load on demand */
  if (database[nr]==NULL) {
    printf("error: database is not present\n");
    return;
  }
  nr_pc=wman+wcrown+bman+bcrown;
  pos_count=db_count(wman,wcrown,bman,bcrown);
  db_cptr=(unsigned char *)malloc(pos_count/5); /* (over-estimated size) */
  if (db_cptr==NULL) {
    printf("fatal: no memory for compressed database\n");
    exit(1);
  }
  dom=countvaluequick(wman,wcrown,bman,bcrown);
  sprintf(db_file,CPR_PATH,id);
  db_start(db_file,nr,pos_count);
  n=init_nextboard(list,type,constraint,min,max,wman,wcrown,bman,bcrown);
  nindex=0;
  set_pieces();
  do {
    index=database_linear_index(white);
    score=read_value(nr,index);
    for (j=nindex;j<index;j++)
      db_addpos(j,dom,false);       /* fill any gaps */
    if (quiet(white)==false)        /* capture position */
      db_addpos(index,dom,false); 
    else
      db_addpos(index,score,true);
    nindex=index+1;
  } while(nextboard(list,type,constraint,min,max,n-1)==false);
  for (j=nindex;j<pos_count;j++)
    db_addpos(j,dom,false);         /* fill any gap at the end */
  db_finish();
  free(db_cptr);
  db_cptr=NULL;
}

/* compress a database, without capture exclusion */
void raz_compress(int nr,char *db_file)
{
  INDEX index,pos_count;
  int wman,wcrown,bman,bcrown,score;
  char *id;

  database_pieces(nr,&wman,&wcrown,&bman,&bcrown);
  id=database_name(wman,wcrown,bman,bcrown);
  printf("raz_compress %s\n",id);
  if (database[nr]==NULL)
    load_database(nr);    /* load on demand */
  if (database[nr]==NULL) {
    printf("error: database is not present\n");
    return;
  }
  nr_pc=wman+wcrown+bman+bcrown;
  pos_count=db_count(wman,wcrown,bman,bcrown);
  db_cptr=database[nr];
  db_start(db_file,nr,pos_count);
  for(index=0;index<pos_count;index++) {
    score=read_value(nr,index);
    if (score==DB_UNKNOWN) score=DB_DRAW;
    db_addpos(index,score,true);
  }
  db_finish();
  free(database[nr]);
  database[nr]=NULL;
}

int database_retrieve_value(int color,int wman,int wcrown,int bman,int bcrown)
/* Retrieves the value of a database entry. */
/* Returns a value from the perspective of 'color'. */
{
  int nr;

  /* no pieces: return LOSE */
  if (color==white) if (wman==0 && wcrown==0) return(DB_LOSE);
  if (color==black) if (bman==0 && bcrown==0) return(DB_LOSE);
  nr=database_nr(color,wman,wcrown,bman,bcrown);
  if (database[nr]==NULL)
    load_database(nr);    /* load on demand */
  if (database[nr]==NULL) {
    printf("error: database is not present\n");
    return(DB_UNKNOWN);
  }
  return read_value(nr,database_linear_index(color));
}

void store_value(int nr,INDEX index,int score)
{
  INDEX dindex;
  unsigned char *db;

  db=database[nr];
  dindex=index/4;
  switch(index-4*dindex) {
  case 0: db[dindex]=(db[dindex]&0x3f)+(score<<6); break;
  case 1: db[dindex]=(db[dindex]&0xcf)+(score<<4); break;
  case 2: db[dindex]=(db[dindex]&0xf3)+(score<<2); break;
  case 3: db[dindex]=(db[dindex]&0xfc)+(score<<0); break;
  }
}

int determine_value(int nr,INDEX index)
/* Determines and sets, if possible, a value in database 'nr', */
/* returns the value. */
{
  int score,nmoves,m,found=DB_UNKNOWN,known,best;

  best=DB_LOSE; known=true;
  nmoves=move_list(0,white);
  for(m=0;m<nmoves;m++) {
    do_move(movelist[0][m]);
    score=database_retrieve_value(black,pieces[white|man],pieces[white|crown],
                                        pieces[black|man],pieces[black|crown]);
    undo_move(movelist[0][m]);
    if (score==DB_UNKNOWN) known=false;
    else if (score==DB_LOSE) {best=DB_WIN; break;}
    else if (score==DB_DRAW) best=DB_DRAW;
  }
  if (best==DB_WIN || known==true) {
    store_value(nr,index,best);
    found=best;
  }
  return(found);
}

INDEX one_iteration_fromqueue(int nr,int wman,int wcrown,int bman,int bcrown,int inq,int outq,int mirror)
{
  INDEX index,found=0;
  int nmoves,m,result;
  int i,j,starttime;
  static char *local;

  starttime=clock();
  init_queue_write(outq);
  init_queue_read(inq);
  init_board();
  for(i=0;i<qsize[inq];i++) {
    local=read_from_queue(inq);
    /* we now have a new board position, with black to move. Undo the
       last (white) move and do a forward search. If results, store into
       the out-queue
       */
    if (mirror==0) for(j=0;j<50;j++) board[map[j]]=local[map[j]];
    else reverse_board(board,local);
    set_pieces();
    nmoves=reverse_move_list(1,white); /* non-captures only ! */
    for(m=0;m<nmoves;m++) {
      do_move(movelist[1][m]); /* we have a potential position, try it ! */
      index=database_linear_index(white);
      if (read_value(nr,index)==DB_UNKNOWN) {
        result=determine_value(nr,index);
        if (result==DB_WIN || result==DB_LOSE) found++;
        if (result==DB_WIN || result==DB_LOSE) add_to_queue(outq);
      }
      undo_move(movelist[1][m]);
    }
  }
  close_queue(inq);
  close_queue(outq);
  printf("*");fflush(stdout);
  sprintf(line,"db:%i%i%i%i, found=%lu, time=%.2f\n",wman,wcrown,bman,bcrown,
          found,(float)(clock()-starttime)/CLOCKS_PER_SEC);
  log(line);
  return(found);
}

INDEX one_iteration(int nr,int wman,int wcrown,int bman,int bcrown,int mode,int outq)
/* Does one iteration in creating database 'nr'. */
/* Returns the number of newly discovered positions */
{
  INDEX index,found=0;
  int n,starttime,result;
  int list[8],type[8],constraint[8],min[8],max[8];

  starttime=clock();
  if (outq>=0) init_queue_write(outq);
  n=init_nextboard(list,type,constraint,min,max,wman,wcrown,bman,bcrown);
  set_pieces();
  do {
    index=database_linear_index(white);
    if (read_value(nr,index)==DB_UNKNOWN &&
        !(mode==CAPTURE && quiet(white)==true)) {
      result=determine_value(nr,index);
      if (result==DB_WIN || result==DB_LOSE) found++;
      if ((result==DB_WIN || result==DB_LOSE) && outq>=0) add_to_queue(outq);
    }
  } while(nextboard(list,type,constraint,min,max,n-1)==false);
  sprintf(line,"db:%i%i%i%i, found=%lu, time=%.2f\n",wman,wcrown,bman,bcrown,
          found,(float)(clock()-starttime)/CLOCKS_PER_SEC);
  log(line);
  if (outq>=0) close_queue(outq);
  return(found);
}

void create_database(int wman,int wcrown,int bman,int bcrown,int level)
{
  INDEX pos_count,found=0,newfound;
  int wm,wc,bm,bc;
  char db_file[2][100],*id;
  int nr[2],p,i;
  int loaded=0,bytesize;
  int q1,q2;
  int mirror;  /* mirror is set to 1 for symmetric databases, 2 otherwise */
  struct stat stbuf;
  
  if (level<0) return;
  /* has it been created ? */
  nr[0]=database_nr(white,wman,wcrown,bman,bcrown);
  nr[1]=database_nr(black,wman,wcrown,bman,bcrown);
  if (database[nr[0]]!=NULL) { /* database is already present */
    return;
  }
  /* first create all required sub-databases */
  for(wm=0;wm<=7;wm++) for(wc=0;wc<=7;wc++)
    for(bm=0;bm<=7;bm++) for(bc=0;bc<=7;bc++)
      if (wm!=wman || bm!=bman || wc!=wcrown || bc!=bcrown) 
        if ((wm+wc)>0 && (bm+bc)>0)
          if (wm<=wman && bm<=bman && wc<=(wman+wcrown) && bc<=(bman+bcrown))
            if ((wm+wc)<=(wman+wcrown) && (bm+bc)<=(bman+bcrown))
              if (wc<=(wcrown+1) && bc<=(bcrown+1))
                create_database(wm,wc,bm,bc,level-1);
  /* symmetric database ? */
  if (nr[0]==nr[1]) mirror=1; else mirror=2;
  /* determine size of the database(s) */
  pos_count=db_count(wman,wcrown,bman,bcrown);
  for(i=0;i<mirror;i++) {
    if (i==0) id=database_name(wman,wcrown,bman,bcrown);
    else id=database_name(bman,bcrown,wman,wcrown);
    sprintf(db_file[i],RAZ_PATH,id);
    /* see if database exists on disk */
    if (my_stat(db_file[i],&stbuf)==-1) {
      close_databases();
      bytesize=(pos_count+3)/4;
      database[nr[i]]=(unsigned char *)malloc(sizeof(unsigned char)*bytesize);
      if (database[nr[i]]==NULL) {
        printf("fatal: no memory for database\n");
        exit(1);
      }
    } else {   
      /* dependent databases will be loaded on demand */
      loaded++;
    }
  }
  if (loaded==mirror) return;
  if (loaded!=0) {
    printf("fatal: weird error on loading\n");
    exit(1);
  }
  /* basic database generation routine */
  /* initialize */
  for(i=0;i<mirror;i++) {
    printf("creating %s, %lu positions\n",db_file[i],pos_count);
    for(p=0;p<pos_count;p++) {
      store_value(nr[i],p,DB_UNKNOWN);
    }
  }
  /* initialise queue system */
  q1=0; q2=1;
  qmax[0]=pos_count/10;
  if (qmax[0]>QMAX) qmax[0]=QMAX;
  qmax[1]=qmax[0];
  qsize[0]=qsize[1]=-1;
  nr_pc=wman+wcrown+bman+bcrown;
  if (mirror==1) {
    /* first pass: capture moves only */
    found=one_iteration(nr[0],wman,wcrown,bman,bcrown,CAPTURE,-1);
    /* second pass: force all positions being seen */
    found+=newfound=one_iteration(nr[0],wman,wcrown,bman,bcrown,NORMAL,q2);
    while(newfound!=0) {
      swapqueues(&q1,&q2);
      if (qsize[q1]<=0) found+=newfound=one_iteration(nr[0],wman,wcrown,bman,bcrown,NORMAL,q2); /* no queue available */
      else found+=newfound=one_iteration_fromqueue(nr[0],wman,wcrown,bman,bcrown,q1,q2,mirror); /* previous computation resulted in a queue */
    }
  } else {
    /* first pass: capture moves only */
    found=one_iteration(nr[0],wman,wcrown,bman,bcrown,CAPTURE,-1);
    found+=one_iteration(nr[1],bman,bcrown,wman,wcrown,CAPTURE,-1);
    /* second pass: force all positions being seen */
    found+=one_iteration(nr[0],wman,wcrown,bman,bcrown,NORMAL,-1);
    found+=one_iteration(nr[1],bman,bcrown,wman,wcrown,NORMAL,q2);
    do {
      swapqueues(&q1,&q2);
      if (qsize[q1]<=0) found+=newfound=one_iteration(nr[0],wman,wcrown,bman,bcrown,NORMAL,q2);
      else found+=newfound=one_iteration_fromqueue(nr[0],wman,wcrown,bman,bcrown,q1,q2,mirror);
      if (newfound==0) break;
      swapqueues(&q1,&q2);
      if (qsize[q1]<=0) found+=newfound=one_iteration(nr[1],bman,bcrown,wman,wcrown,NORMAL,q2);
      else found+=newfound=one_iteration_fromqueue(nr[1],bman,bcrown,wman,wcrown,q1,q2,mirror);
    } while(newfound!=0);
  }
  printf("\na total of %lu positions have been solved\n",found);
  /* write databases */
  for(i=0;i<mirror;i++)
    raz_compress(nr[i],db_file[i]);
}

void debug(int wman,int wcrown,int bman,int bcrown)
/* verify database values */
{
  INDEX index;
  INDEX count[4]={0,0,0,0};
  int nr,i,n,best,known,original,score,nmoves,m;
  int list[8],type[8],constraint[8],min[8],max[8];
  char *id;

  id=database_name(wman,wcrown,bman,bcrown);
  printf("verify %s\n",id);
  nr=database_nr(white,wman,wcrown,bman,bcrown);
  if (database[nr]==NULL)
    load_database(nr);    /* load on demand */
  if (database[nr]==NULL) {
    printf("error: database is not present\n");
    return;
  }
  n=init_nextboard(list,type,constraint,min,max,wman,wcrown,bman,bcrown);
  i=0;
  set_pieces();
  do {
    index=database_linear_index(white);
    original=read_value(nr,index);
    count[original]++;
    best=DB_LOSE; known=true;
    nmoves=move_list(0,white);
    for(m=0;m<nmoves;m++) {
      do_move(movelist[0][m]);
      score=database_retrieve_value(black,pieces[white|man],pieces[white|crown],pieces[black|man],pieces[black|crown]);
      undo_move(movelist[0][m]);
      if (score==DB_UNKNOWN) known=false;
      if (score==DB_LOSE) {best=DB_WIN; break;}
      if (score==DB_DRAW) best=DB_DRAW;
    }
    if (known==false || best!=original) {
      printf("error found! nr:%i ply0 value:%i ply1 value:%i, index:%lu\n",i,original,best,index);
      display_board();
      i++;
    }
  } while(nextboard(list,type,constraint,min,max,n-1)==false);
  id=database_name(wman,wcrown,bman,bcrown);
  sprintf(line,"verified %s, errors:%i\n",id,i);
  log(line);
  sprintf(line, "database %s:   win: %10lu   draw: %10lu   loss: %10lu\n",
         id, count[DB_WIN],count[DB_DRAW],count[DB_LOSE]);
  log(line);
  if (count[DB_UNKNOWN]!=0)
    printf("error: database has %lu 'unknown' values\n",count[DB_UNKNOWN]);
}

void verify_all(wman,wcrown,bman,bcrown,level)
{
  int wm,wc,bm,bc,nr[2];

  if (level<0) return;
  nr[0]=database_nr(white,wman,wcrown,bman,bcrown);
  nr[1]=database_nr(black,wman,wcrown,bman,bcrown);
  if (verified[nr[0]]==true) return;
  /* first verify all sub-databases */
  for(wm=0;wm<=7;wm++) for(wc=0;wc<=7;wc++)
    for(bm=0;bm<=7;bm++) for(bc=0;bc<=7;bc++)
      if (wm!=wman || bm!=bman || wc!=wcrown || bc!=bcrown) 
        if ((wm+wc)>0 && (bm+bc)>0)
          if (wm<=wman && bm<=bman && wc<=(wman+wcrown) && bc<=(bman+bcrown))
            if ((wm+wc)<=(wman+wcrown) && (bm+bc)<=(bman+bcrown))
              verify_all(wm,wc,bm,bc,level-1);
  debug(wman,wcrown,bman,bcrown);
  if (nr[0]!=nr[1]) debug(bman,bcrown,wman,wcrown);
  verified[nr[0]]=verified[nr[1]]=true;
}

int database_value(int color,int wman,int wcrown,int bman,int bcrown)
/* checks if board[] is in a loaded database. If it is, it returns
   one of:WIN,DRAW,LOSE or UNKNOWN, seen from the white party. Inputs:
   color: player to move
   wman,wcrown,bman,bcrown: the number of these pieces on the board.
   In general these are known, so it speeds up the access.
*/
{
  INDEX index;
  int score,nr;

  if (use_db==false) return(UNKNOWN);
  nr=database_nr(color,wman,wcrown,bman,bcrown);
  if (database[nr]==NULL) return(UNKNOWN);
  index=database_linear_index(color);
  score=read_value(nr,index);
  ndat++; db_usage[nr]++;
  if (color==white) {
        if (score==DB_DRAW) return(DRAW);
        if (score==DB_WIN) return(WIN);
        if (score==DB_LOSE) return(LOSE);
  } else {
        if (score==DB_DRAW) return(DRAW);
        if (score==DB_WIN) return(LOSE);
        if (score==DB_LOSE) return(WIN);
  }
  dbfail++;
  return(UNKNOWN);
}

int read_all_databases(void)
/* reads in all available databases defined in the file 'DB_INDEX_FILE' */
{
  FILE *in;
  char id[32],include;
  int wman,wcrown,bman,bcrown,ndb,ii,n,size=0;

  in=my_fopen(DB_INDEX_FILE,"r");
  if (in==NULL) {
    printf("error: database index file not found\n");
    return 0;
  }
  /* get number of databases to be loaded */
  if (fscanf(in,"%i\n",&ndb)!=1) ndb=0;
  /* read all selected databases */
  for(ii=0;ii<ndb;ii++) {
    if (fscanf(in,"%s %c",id,&include)!=2) continue;
    if (include=='n') continue;
    n=wman=bman=wcrown=bcrown=0;
    while(id[n]!='v'&&id[n]!=0) {
      if (id[n]=='X') wcrown++;
      if (id[n]=='O') wman++;
      n++;
    }
    while(id[n]!=0) {
      if (id[n]=='X') bcrown++;
      if (id[n]=='O') bman++;
      n++;
    }
    printf("opening database %s:%i%i%i%i\r",
           database_name(wman,wcrown,bman,bcrown),wman,wcrown,bman,bcrown);
    fflush(stdout);
    n=database_nr(white,wman,wcrown,bman,bcrown);
    size+=load_database(n);
    loadeddb[n]=true;
    printf("                                                          \r");
    fflush(stdout);
  }
  fclose(in);
  return size;
}  
  
void countvalue(int wman,int wcrown,int bman,int bcrown,int exclcapt)
/* prints result-counts of a database.  */
{
  INDEX count[4]={0,0,0,0};
  int nr,n;
  int list[8],type[8],constraint[8],min[8],max[8];
  char *id;

  id=database_name(wman,wcrown,bman,bcrown);
  nr=database_nr(white,wman,wcrown,bman,bcrown);
  if (database[nr]==NULL)
    load_database(nr);    /* load on demand */
  if (database[nr]==NULL) {
    printf("error: database is not present\n");
    return;
  }
  n=init_nextboard(list,type,constraint,min,max,wman,wcrown,bman,bcrown);
  set_pieces();
  do {
    if (exclcapt && quiet(white)==false) /* capture position */
      continue;
    count[read_value(nr,database_linear_index(white))]++;
  } while(nextboard(list,type,constraint,min,max,n-1)==false);
  printf("database %s:   win: %10lu   draw: %10lu   loss: %10lu\n",
         database_name(wman,wcrown,bman,bcrown),
         count[DB_WIN],count[DB_DRAW],count[DB_LOSE]);
  if (count[DB_UNKNOWN]!=0)
    printf("error: database has %lu 'unknown' values\n",count[DB_UNKNOWN]);
  return;
}

void count_all(int exclcapt)
{
  int i,wman,wcrown,bman,bcrown;

  for(i=0; i<4096; i++) {
    if (loadeddb[i] == true) {
      database_pieces(i,&wman,&wcrown,&bman,&bcrown);
      countvalue(wman,wcrown,bman,bcrown,exclcapt);
    }
  }
}

void size_db(int pieces)
{
  INDEX pos_count;
  int i,wman,wcrown,bman,bcrown;

  for(i=0; i<4096; i++) {
    database_pieces(i,&wman,&wcrown,&bman,&bcrown);
    if (wman+wcrown>0&&bman+bcrown>0&&wman+wcrown+bman+bcrown==pieces) {
      pos_count=db_count(wman,wcrown,bman,bcrown);
      printf("database %s:   size: %10lu   bytes: %10lu\n",
         database_name(wman,wcrown,bman,bcrown), pos_count, (pos_count+3)/4);
    }
  }
}

void debug_all(void)
{
  int i,wman,wcrown,bman,bcrown;

  for(i=0; i<4096; i++) {
    if (loadeddb[i] == true) {
      database_pieces(i,&wman,&wcrown,&bman,&bcrown);
      debug(wman,wcrown,bman,bcrown);
    }
  }
}

/* read in a compressed database file */
int read_compr_file(int wman,int wcrown,int bman,int bcrown)
{
  char db_file[100],*id;
  int size,nr;
  FILE *db;
  struct stat stbuf;

  id=database_name(wman,wcrown,bman,bcrown);
  sprintf(db_file,CPR_PATH,id);
  if (my_stat(db_file,&stbuf)==-1) {
    printf("error: compressed database %s not found\n",db_file);
    return false;
  }
  size=stbuf.st_size;
  nr=database_nr(white,wman,wcrown,bman,bcrown);
  cpr_database[nr]=(unsigned char *)malloc(sizeof(unsigned char)*size);
  if (cpr_database[nr]==NULL) {
    printf("error: no memory for database\n");
    return false;
  }
  db=fopen(db_file,"rb");
  if (db==NULL) {
    printf("error: compressed database %s not found\n",db_file);
    free(cpr_database[nr]);
    cpr_database[nr]=NULL;
    return false;
  }
  printf("opening compressed database %s:%i%i%i%i, size=%i\n",
         db_file,wman,wcrown,bman,bcrown,size);
  if (fread(cpr_database[nr],sizeof(unsigned char),size,db) != size) {
    printf("error: compressed database %s read failure\n",db_file);
    free(cpr_database[nr]);
    cpr_database[nr]=NULL;
    return false;
  }
  fclose(db);
  return true;
}

/* verify a compressed database */
void db_cpr_verify(int wman,int wcrown,int bman,int bcrown)
{
  INDEX index;
  int nr,n,score,cmprval;
  int list[8],type[8],constraint[8],min[8],max[8];
  char *id;

  id=database_name(wman,wcrown,bman,bcrown);
  printf("db_cpr_verify %s\n",id);
  nr=database_nr(white,wman,wcrown,bman,bcrown);
  if (database[nr]==NULL)
    load_database(nr);    /* load on demand */
  if (database[nr]==NULL) {
    printf("error: database is not present\n");
    return;
  }
  if (cpr_database[nr]==NULL)
    if (!read_compr_file(wman,wcrown,bman,bcrown))
      return;
  n=init_nextboard(list,type,constraint,min,max,wman,wcrown,bman,bcrown);
  set_pieces();
  do {
    index=database_linear_index(white);
    score=read_value(nr,index);
    if (quiet(white)==false) /* capture position */
      continue;
    cmprval=read_compr(nr,index);
    if (cmprval!=score) {
      printf("compressed db value differs! index=%lu score=%d cmprval=%d\n",
             index,score,cmprval);
      display_board();
      exit(1);
    }
  } while(nextboard(list,type,constraint,min,max,n-1)==false);
  free(cpr_database[nr]);
  cpr_database[nr]=NULL;
}

void compress_all(void)
{
  int i,wman,wcrown,bman,bcrown;

  for(i=0; i<4096; i++) {
    if (loadeddb[i] == true) {
      close_databases();    /* make room for target memory block */
      database_pieces(i,&wman,&wcrown,&bman,&bcrown);
      db_compress(wman,wcrown,bman,bcrown);
#if 0
      db_cpr_verify(wman,wcrown,bman,bcrown);
#endif
    }
  }
}

int db_main(int argc, char *argv[])
{
  int i;

  log("Started\n");
  init_var();
  init_databases();

  for (i=1; i<argc; i++) {
    if        (strcmp(argv[i],"count")==0) {
      read_all_databases();
      count_all(false);
    } else if (strcmp(argv[i],"countex")==0) { /* exclude capture positions */
      read_all_databases();
      count_all(true);
    } else if (strcmp(argv[i],"compress")==0) {
      read_all_databases();
      compress_all();
    } else if (strcmp(argv[i],"verify")==0) {
      read_all_databases();
      debug_all();
    } else if (strcmp(argv[i],"size")==0) {
      i++;
      if (i>=argc) break;
      size_db(atoi(argv[i]));
    } else if (strcmp(argv[i],"create")==0) {
      i+=4;
      if (i>=argc) break;
#ifdef _WIN32
      mkdir("/tmp"); /* just in case it doesn't exist on current partition */
#endif
      create_database(atoi(argv[i-3]),atoi(argv[i-2]),
                      atoi(argv[i-1]),atoi(argv[i]),10);
    }
  }
  log("Finished\n\n");
  return(0);
}
